import typing
from importlib.metadata import version
from types import TracebackType
from typing import Any, Final, Optional

from packaging.version import Version
from typing_extensions import deprecated, override

from that_depends import container_context
from that_depends.providers.context_resources import ContextScope, SupportsContext
from that_depends.utils import UNSET, Unset, is_set


_FASTSTREAM_MODULE_NAME: Final[str] = "faststream"
_FASTSTREAM_VERSION: Final[str] = version(_FASTSTREAM_MODULE_NAME)
if Version(_FASTSTREAM_VERSION) >= Version("0.6.0"):  # pragma: no cover
    from faststream import BaseMiddleware, ContextRepo
    from faststream._internal.types import AnyMsg

    class DIContextMiddleware(BaseMiddleware):
        """Initializes the container context for faststream brokers."""

        def __init__(
            self,
            *context_items: SupportsContext[Any],
            msg: AnyMsg | None = None,
            context: Optional["ContextRepo"] = None,
            global_context: dict[str, Any] | Unset = UNSET,
            scope: ContextScope | Unset = UNSET,
        ) -> None:
            """Initialize the container context middleware.

            Args:
                *context_items (SupportsContext[Any]): Context items to initialize.
                msg (Any): Message object.
                context (ContextRepo): Context repository.
                global_context (dict[str, Any] | Unset): Global context to initialize the container.
                scope (ContextScope | Unset): Context scope to initialize the container.

            """
            super().__init__(msg, context=context)  # type: ignore[arg-type]
            self._context: container_context | None = None
            self._context_items = set(context_items)
            self._global_context = global_context
            self._scope = scope

        @override
        async def on_receive(self) -> None:
            self._context = container_context(
                *self._context_items,
                scope=self._scope if is_set(self._scope) else None,
                global_context=self._global_context if is_set(self._global_context) else None,
            )
            await self._context.__aenter__()

        @override
        async def after_processed(
            self,
            exc_type: type[BaseException] | None = None,
            exc_val: BaseException | None = None,
            exc_tb: Optional["TracebackType"] = None,
        ) -> bool | None:
            if self._context is not None:
                await self._context.__aexit__(exc_type, exc_val, exc_tb)
            return None

        def __call__(self, msg: Any = None, **kwargs: Any) -> "DIContextMiddleware":  # noqa: ANN401
            """Create an instance of DIContextMiddleware.

            Args:
                msg (Any): Message object.
                **kwargs: Additional keyword arguments.

            Returns:
                DIContextMiddleware: A new instance of DIContextMiddleware.

            """
            context = kwargs.get("context")

            return DIContextMiddleware(
                *self._context_items,
                msg=msg,
                context=context,
                scope=self._scope,
                global_context=self._global_context,
            )
else:  # pragma: no cover
    from faststream import BaseMiddleware

    @deprecated("Will be removed with faststream v1")
    class DIContextMiddleware(BaseMiddleware):  # type: ignore[no-redef]
        """Initializes the container context for faststream brokers."""

        def __init__(
            self,
            *context_items: SupportsContext[Any],
            global_context: dict[str, Any] | Unset = UNSET,
            scope: ContextScope | Unset = UNSET,
        ) -> None:
            """Initialize the container context middleware.

            Args:
                *context_items (SupportsContext[Any]): Context items to initialize.
                global_context (dict[str, Any] | Unset): Global context to initialize the container.
                scope (ContextScope | Unset): Context scope to initialize the container.

            """
            super().__init__()  # type: ignore[call-arg]
            self._context: container_context | None = None
            self._context_items = set(context_items)
            self._global_context = global_context
            self._scope = scope

        @override
        async def on_receive(self) -> None:
            self._context = container_context(
                *self._context_items,
                scope=self._scope if is_set(self._scope) else None,
                global_context=self._global_context if is_set(self._global_context) else None,
            )
            await self._context.__aenter__()

        @override
        async def after_processed(
            self,
            exc_type: type[BaseException] | None = None,
            exc_val: BaseException | None = None,
            exc_tb: Optional["TracebackType"] = None,
        ) -> bool | None:
            if self._context is not None:
                await self._context.__aexit__(exc_type, exc_val, exc_tb)
            return None

        def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> "DIContextMiddleware":  # noqa: ARG002, ANN401
            """Create an instance of DIContextMiddleware."""
            return DIContextMiddleware(*self._context_items, scope=self._scope, global_context=self._global_context)



import importlib
import re
import typing
from typing import Any, TypeVar, cast, overload

import typing_extensions

from that_depends import ContextScope
from that_depends.providers import AbstractProvider
from that_depends.providers.context_resources import CT, SupportsContext
from that_depends.providers.mixin import SupportsTeardown


T_co = TypeVar("T_co", covariant=True)

_IMPORT_STRING_REGEX = re.compile(r"^[a-zA-Z_][a-zA-Z0-9_]*(\.[a-zA-Z_][a-zA-Z0-9_]*)*$")


class LazyProvider(SupportsTeardown, SupportsContext[Any], AbstractProvider[Any]):
    """Lazily imports and provides a provider from a module."""

    @overload
    def __init__(self, *, module_string: str, provider_string: str) -> None: ...

    @overload
    def __init__(self, import_string: str, /) -> None: ...

    def __init__(
        self,
        import_string: str | None = None,
        module_string: str | None = None,
        provider_string: str | None = None,
    ) -> None:
        """Initialize a LazyProvider instance.

        Args:
            module_string: path to module to import from.
            provider_string: path to provider within module.
            import_string: path to provider including module and attributes.

        """
        super().__init__()
        if (import_string is not None) == (module_string is not None and provider_string is not None):
            msg = (
                "You must provide either import_string OR both module_string AND provider_string, "
                "but not both or neither."
            )
            raise ValueError(msg)

        self._module_string = module_string
        self._provider_string = provider_string
        self._import_string = import_string
        self._check_strings()
        self._provider: AbstractProvider[Any] | None = None

    @typing_extensions.override
    def get_scope(self) -> ContextScope | None:
        provider = self._get_provider()
        if isinstance(provider, SupportsContext):
            return provider.get_scope()
        msg = "Underlying provider does not support context scopes"
        raise NotImplementedError(msg)

    @typing_extensions.override
    def context_async(self, force: bool = False) -> typing.AsyncContextManager[CT]:
        provider = self._get_provider()
        if isinstance(provider, SupportsContext):
            return provider.context_async(force)
        msg = "Underlying provider does not support context management"
        raise NotImplementedError(msg)

    @typing_extensions.override
    def context_sync(self, force: bool = False) -> typing.ContextManager[CT]:
        provider = self._get_provider()
        if isinstance(provider, SupportsContext):
            return provider.context_sync(force)
        msg = "Underlying provider does not support context management"
        raise NotImplementedError(msg)

    @typing_extensions.override
    def supports_context_sync(self) -> bool:
        provider = self._get_provider()
        return isinstance(provider, SupportsContext) and provider.supports_context_sync()

    @typing_extensions.override
    async def tear_down(self, propagate: bool = True) -> None:
        provider = self._get_provider()
        if isinstance(provider, SupportsTeardown):
            await provider.tear_down(propagate=propagate)
        else:
            msg = "Underlying provider does not support tear down."
            raise NotImplementedError(msg)

    @typing_extensions.override
    def tear_down_sync(self, propagate: bool = True, raise_on_async: bool = True) -> None:
        provider = self._get_provider()
        if isinstance(provider, SupportsTeardown):
            provider.tear_down_sync(propagate=propagate, raise_on_async=raise_on_async)
        else:
            msg = "Underlying provider does not support tear down."
            raise NotImplementedError(msg)

    def _check_strings(self) -> None:
        if self._import_string is not None and not _IMPORT_STRING_REGEX.match(self._import_string):
            msg = f"Invalid import_string '{self._import_string}'"
            raise ValueError(msg)
        if self._module_string is not None and not _IMPORT_STRING_REGEX.match(self._module_string):
            msg = f"Invalid module_string '{self._module_string}'"
            raise ValueError(msg)
        if self._provider_string is not None and not _IMPORT_STRING_REGEX.match(self._provider_string):
            msg = f"Invalid provider_string '{self._provider_string}'"
            raise ValueError(msg)

    def _get_provider(self) -> AbstractProvider[Any]:
        if self._provider:
            return self._provider
        if self._import_string is not None:
            parts = self._import_string.split(".")
            for i in range(len(parts), 0, -1):
                module_name = ".".join(parts[:i])
                try:
                    module = importlib.import_module(module_name)
                    attrs = parts[i:]
                    break
                except ImportError:
                    continue
            else:
                msg = f"Cannot import any module from '{self._import_string}'"
                raise ImportError(msg)
        else:
            if self._module_string is None or self._provider_string is None:
                msg = "Invalid state: module_string and provider_string must be set"
                raise RuntimeError(msg)
            module = importlib.import_module(self._module_string)
            attrs = self._provider_string.split(".")
        provider = module
        for attr in attrs:
            provider = getattr(provider, attr)
        self._provider = cast(AbstractProvider[Any], provider)
        return self._provider

    @typing_extensions.override
    async def resolve(self) -> Any:
        provider = self._get_provider()
        return await provider.resolve()

    @typing_extensions.override
    def resolve_sync(self) -> Any:
        provider = self._get_provider()
        return provider.resolve_sync()

    @typing_extensions.override
    def override_sync(
        self, mock: object, tear_down_children: bool = False, propagate: bool = True, raise_on_async: bool = False
    ) -> None:
        provider = self._get_provider()
        provider.override_sync(mock, tear_down_children, propagate, raise_on_async)

    @typing_extensions.override
    async def override(self, mock: object, tear_down_children: bool = False, propagate: bool = True) -> None:
        provider = self._get_provider()
        await provider.override(mock, tear_down_children, propagate)

    @typing_extensions.override
    async def reset_override(self, tear_down_children: bool = False, propagate: bool = True) -> None:
        provider = self._get_provider()
        await provider.reset_override(tear_down_children, propagate)

    @typing_extensions.override
    def reset_override_sync(
        self, tear_down_children: bool = False, propagate: bool = True, raise_on_async: bool = False
    ) -> None:
        provider = self._get_provider()
        provider.reset_override_sync(tear_down_children, propagate, raise_on_async)

    @typing_extensions.override
    def __getattr__(self, attr_name: str) -> typing.Any:
        provider = self._get_provider()
        return getattr(provider, attr_name)
